@@ -28,7 +28,6 @@
  * POSSIBILITY OF SUCH DAMAGE.
  */
 #include <util.h>
-/*@unused@*/ RCSID("$Id: cv-symline.c 2130 2008-10-07 05:38:11Z peter $");
 
 #include <libyasm.h>
 
@@ -128,13 +127,16 @@ typedef struct cv8_lineset {
     size_t num_pairs;
 } cv8_lineset;
 
+/* Note: Due to line number sorting requirements (by section offset it seems)
+ *       one file may need more than one record per section. */
 typedef struct cv8_lineinfo {
     STAILQ_ENTRY(cv8_lineinfo) link;
     const cv_filename *fn;      /* filename associated with line numbers */
     yasm_section *sect;         /* section line numbers are for */
     yasm_symrec *sectsym;       /* symbol for beginning of sect */
     unsigned long num_linenums;
-    STAILQ_HEAD(, cv8_lineset) linesets;
+    int first_in_sect;          /* First lineinfo for this section. */
+    STAILQ_HEAD(cv8_lineset_head, cv8_lineset) linesets;
 } cv8_lineinfo;
 
 /* Symbols use a bit of meta-programming to encode formats: each character
@@ -163,7 +165,7 @@ static void cv8_symhead_bc_print(const void *contents, FILE *f,
 static int cv8_symhead_bc_calc_len
     (yasm_bytecode *bc, yasm_bc_add_span_func add_span, void *add_span_data);
 static int cv8_symhead_bc_tobytes
-    (yasm_bytecode *bc, unsigned char **bufp, void *d,
+    (yasm_bytecode *bc, unsigned char **bufp, unsigned char *bufstart, void *d,
      yasm_output_value_func output_value,
      /*@null@*/ yasm_output_reloc_func output_reloc);
 
@@ -173,7 +175,7 @@ static void cv8_fileinfo_bc_print(const void *contents, FILE *f,
 static int cv8_fileinfo_bc_calc_len
     (yasm_bytecode *bc, yasm_bc_add_span_func add_span, void *add_span_data);
 static int cv8_fileinfo_bc_tobytes
-    (yasm_bytecode *bc, unsigned char **bufp, void *d,
+    (yasm_bytecode *bc, unsigned char **bufp, unsigned char *bufstart, void *d,
      yasm_output_value_func output_value,
      /*@null@*/ yasm_output_reloc_func output_reloc);
 
@@ -183,7 +185,7 @@ static void cv8_lineinfo_bc_print(const void *contents, FILE *f,
 static int cv8_lineinfo_bc_calc_len
     (yasm_bytecode *bc, yasm_bc_add_span_func add_span, void *add_span_data);
 static int cv8_lineinfo_bc_tobytes
-    (yasm_bytecode *bc, unsigned char **bufp, void *d,
+    (yasm_bytecode *bc, unsigned char **bufp, unsigned char *bufstart, void *d,
      yasm_output_value_func output_value,
      /*@null@*/ yasm_output_reloc_func output_reloc);
 
@@ -192,7 +194,7 @@ static void cv_sym_bc_print(const void *contents, FILE *f, int indent_level);
 static int cv_sym_bc_calc_len
     (yasm_bytecode *bc, yasm_bc_add_span_func add_span, void *add_span_data);
 static int cv_sym_bc_tobytes
-    (yasm_bytecode *bc, unsigned char **bufp, void *d,
+    (yasm_bytecode *bc, unsigned char **bufp, unsigned char *bufstart, void *d,
      yasm_output_value_func output_value,
      /*@null@*/ yasm_output_reloc_func output_reloc);
 
@@ -419,7 +421,7 @@ typedef struct cv_line_info {
     yasm_linemap *linemap;
     yasm_errwarns *errwarns;
     unsigned int num_lineinfos;
-    STAILQ_HEAD(, cv8_lineinfo) cv8_lineinfos;
+    STAILQ_HEAD(cv8_lineinfo_head, cv8_lineinfo) cv8_lineinfos;
     /*@null@*/ cv8_lineinfo *cv8_cur_li;
     /*@null@*/ cv8_lineset *cv8_cur_ls;
 } cv_line_info;
@@ -444,23 +446,9 @@ cv_generate_line_bc(yasm_bytecode *bc, /*@null@*/ void *d)
         || strcmp(filename, info->cv8_cur_li->fn->filename) != 0) {
         yasm_bytecode *sectbc;
         char symname[8];
+        int first_in_sect = !info->cv8_cur_li;
 
-        /* first see if we already have a lineinfo that is for this section and
-         * filename
-         */
-        STAILQ_FOREACH(info->cv8_cur_li, &info->cv8_lineinfos, link) {
-            if (sect == info->cv8_cur_li->sect
-                && strcmp(filename, info->cv8_cur_li->fn->filename) == 0)
-                break;
-        }
-
-        if (info->cv8_cur_li) {
-            info->cv8_cur_ls = STAILQ_LAST(&info->cv8_cur_li->linesets,
-                                           cv8_lineset, link);
-            goto done;          /* found one */
-        }
-
-        /* Nope; find file */
+        /* Find file */
         for (i=0; i<dbgfmt_cv->filenames_size; i++) {
             if (strcmp(filename, dbgfmt_cv->filenames[i].filename) == 0)
                 break;
@@ -472,6 +460,7 @@ cv_generate_line_bc(yasm_bytecode *bc, /*@null@*/ void *d)
         info->cv8_cur_li = yasm_xmalloc(sizeof(cv8_lineinfo));
         info->cv8_cur_li->fn = &dbgfmt_cv->filenames[i];
         info->cv8_cur_li->sect = sect;
+        info->cv8_cur_li->first_in_sect = first_in_sect;
         sectbc = yasm_section_bcs_first(sect);
         if (sectbc->symrecs && sectbc->symrecs[0])
             info->cv8_cur_li->sectsym = sectbc->symrecs[0];
@@ -486,7 +475,6 @@ cv_generate_line_bc(yasm_bytecode *bc, /*@null@*/ void *d)
         STAILQ_INSERT_TAIL(&info->cv8_lineinfos, info->cv8_cur_li, link);
         info->cv8_cur_ls = NULL;
     }
-done:
 
     /* build new lineset if necessary */
     if (!info->cv8_cur_ls || info->cv8_cur_ls->num_pairs >= 126) {
@@ -623,13 +611,19 @@ yasm_cv__generate_symline(yasm_object *object, yasm_linemap *linemap,
                                   cv_generate_line_section);
 
     /* Output line numbers for sections */
+    head = NULL;
     STAILQ_FOREACH(li, &info.cv8_lineinfos, link) {
-        head = cv8_add_symhead(info.debug_symline, CV8_LINE_NUMS, 0);
+        if (li->first_in_sect) {
+            if (head)
+                cv8_set_symhead_end(head, yasm_section_bcs_last(info.debug_symline));
+            head = cv8_add_symhead(info.debug_symline, CV8_LINE_NUMS, 0);
+        }
         bc = yasm_bc_create_common(&cv8_lineinfo_bc_callback, li, 0);
-        bc->len = 24+li->num_linenums*8;
+        bc->len = (li->first_in_sect ? 24 : 12) + li->num_linenums*8;
         yasm_cv__append_bc(info.debug_symline, bc);
-        cv8_set_symhead_end(head, yasm_section_bcs_last(info.debug_symline));
     }
+    if (head)
+        cv8_set_symhead_end(head, yasm_section_bcs_last(info.debug_symline));
 
     /* Already aligned 4 */
 
@@ -638,9 +632,12 @@ yasm_cv__generate_symline(yasm_object *object, yasm_linemap *linemap,
     /* add object and compile flag first */
     cv8_add_sym_objname(info.debug_symline,
                         yasm__abspath(object->obj_filename));
-    cv8_add_sym_compile(object, info.debug_symline,
-                        yasm__xstrdup(PACKAGE_NAME " " PACKAGE_INTVER "."
-                                      PACKAGE_BUILD));
+    if (getenv("YASM_TEST_SUITE"))
+        cv8_add_sym_compile(object, info.debug_symline,
+                            yasm__xstrdup("yasm HEAD"));
+    else
+        cv8_add_sym_compile(object, info.debug_symline,
+                            yasm__xstrdup(PACKAGE_STRING));
     /* then iterate through symbol table */
     yasm_symtab_traverse(object->symtab, &info, cv_generate_sym);
     cv8_set_symhead_end(head, yasm_section_bcs_last(info.debug_symline));
@@ -724,7 +721,8 @@ cv8_symhead_bc_calc_len(yasm_bytecode *bc, yasm_bc_add_span_func add_span,
 }
 
 static int
-cv8_symhead_bc_tobytes(yasm_bytecode *bc, unsigned char **bufp, void *d,
+cv8_symhead_bc_tobytes(yasm_bytecode *bc, unsigned char **bufp,
+                       unsigned char *bufstart, void *d,
                        yasm_output_value_func output_value,
                        yasm_output_reloc_func output_reloc)
 {
@@ -797,7 +795,8 @@ cv8_fileinfo_bc_calc_len(yasm_bytecode *bc, yasm_bc_add_span_func add_span,
 }
 
 static int
-cv8_fileinfo_bc_tobytes(yasm_bytecode *bc, unsigned char **bufp, void *d,
+cv8_fileinfo_bc_tobytes(yasm_bytecode *bc, unsigned char **bufp,
+                        unsigned char *bufstart, void *d,
                         yasm_output_value_func output_value,
                         yasm_output_reloc_func output_reloc)
 {
@@ -864,7 +863,8 @@ cv8_lineinfo_bc_calc_len(yasm_bytecode *bc, yasm_bc_add_span_func add_span,
 }
 
 static int
-cv8_lineinfo_bc_tobytes(yasm_bytecode *bc, unsigned char **bufp, void *d,
+cv8_lineinfo_bc_tobytes(yasm_bytecode *bc, unsigned char **bufp,
+                        unsigned char *bufstart, void *d,
                         yasm_output_value_func output_value,
                         yasm_output_reloc_func output_reloc)
 {
@@ -875,21 +875,25 @@ cv8_lineinfo_bc_tobytes(yasm_bytecode *bc, unsigned char **bufp, void *d,
     unsigned long i;
     cv8_lineset *ls;
 
-    /* start offset and section */
-    cv_out_sym(li->sectsym, 0, bc, &buf, d, output_value);
+    if (li->first_in_sect) {
+        /* start offset and section */
+        cv_out_sym(li->sectsym, (unsigned long)(buf - bufstart), bc, &buf,
+                   d, output_value);
 
-    /* Two bytes of pad/alignment */
-    YASM_WRITE_8(buf, 0);
-    YASM_WRITE_8(buf, 0);
+        /* Two bytes of pad/alignment */
+        YASM_WRITE_8(buf, 0);
+        YASM_WRITE_8(buf, 0);
 
-    /* Section length covered by line number info */
-    cval = yasm_calc_bc_dist(yasm_section_bcs_first(li->sect),
-                             yasm_section_bcs_last(li->sect));
-    yasm_arch_intnum_tobytes(object->arch, cval, buf, 4, 32, 0, bc, 0);
-    buf += 4;
+        /* Section length covered by line number info */
+        cval = yasm_calc_bc_dist(yasm_section_bcs_first(li->sect),
+                                 yasm_section_bcs_last(li->sect));
+        yasm_arch_intnum_tobytes(object->arch, cval, buf, 4, 32, 0, bc, 0);
+        yasm_intnum_destroy(cval);
+        buf += 4;
+    }
 
     /* Offset of source file in info table */
-    yasm_intnum_set_uint(cval, li->fn->info_off);
+    cval = yasm_intnum_create_uint(li->fn->info_off);
     yasm_arch_intnum_tobytes(object->arch, cval, buf, 4, 32, 0, bc, 0);
     buf += 4;
 
@@ -1019,7 +1023,8 @@ cv_sym_bc_calc_len(yasm_bytecode *bc, yasm_bc_add_span_func add_span,
 }
 
 static int
-cv_sym_bc_tobytes(yasm_bytecode *bc, unsigned char **bufp, void *d,
+cv_sym_bc_tobytes(yasm_bytecode *bc, unsigned char **bufp,
+                  unsigned char *bufstart, void *d,
                   yasm_output_value_func output_value,
                   yasm_output_reloc_func output_reloc)
 {
@@ -1061,7 +1066,7 @@ cv_sym_bc_tobytes(yasm_bytecode *bc, unsigned char **bufp, void *d,
                 break;
             case 'Y':
                 cv_out_sym((yasm_symrec *)cvs->args[arg++].p,
-                           (unsigned long)(buf-(*bufp)), bc, &buf, d,
+                           (unsigned long)(buf-bufstart), bc, &buf, d,
                            output_value);
                 break;
             case 'T':
