@@ -1,3 +1,7 @@
+#ifdef _WIN32
+#include <windows.h>
+#include <io.h>
+#endif
 #include <stdlib.h>
 #include <string.h>
 #include <ctype.h>
@@ -6,6 +10,57 @@
 #include "tools/re2c/dfa.h"
 #include "tools/re2c/parse.h"
 
+#ifdef _WIN32
+/* tmpfile() replacment for Windows.
+ *
+ * On Windows tmpfile() creates the file in the root directory. This
+ * may fail due to unsufficient privileges.
+ */
+static FILE *
+win32_tmpfile (void)
+{
+    DWORD path_len;
+    WCHAR path_name[MAX_PATH + 1];
+    WCHAR file_name[MAX_PATH + 1];
+    HANDLE handle;
+    int fd;
+    FILE *fp;
+
+    path_len = GetTempPathW (MAX_PATH, path_name);
+    if (path_len <= 0 || path_len >= MAX_PATH)
+	return NULL;
+
+    if (GetTempFileNameW (path_name, L"ps_", 0, file_name) == 0)
+	return NULL;
+
+    handle = CreateFileW (file_name,
+			 GENERIC_READ | GENERIC_WRITE,
+			 0,
+			 NULL,
+			 CREATE_ALWAYS,
+			 FILE_ATTRIBUTE_NORMAL | FILE_FLAG_DELETE_ON_CLOSE,
+			 NULL);
+    if (handle == INVALID_HANDLE_VALUE) {
+	DeleteFileW (file_name);
+	return NULL;
+    }
+
+    fd = _open_osfhandle((intptr_t) handle, 0);
+    if (fd < 0) {
+	CloseHandle (handle);
+	return NULL;
+    }
+
+    fp = fdopen(fd, "w+b");
+    if (fp == NULL) {
+	_close(fd);
+	return NULL;
+    }
+
+    return fp;
+}
+#endif
+
 static void useLabel(size_t value) {
     while (value >= vUsedLabelAlloc) {
 	vUsedLabels = realloc(vUsedLabels, vUsedLabelAlloc * 2);
@@ -146,11 +201,11 @@ void BitMap_gen(FILE *o, unsigned int lb, unsigned int ub){
 	unsigned int n = ub - lb;
 	unsigned int i;
 	unsigned char *bm = malloc(sizeof(unsigned char)*n);
-	memset(bm, 0, n);
 	fputs("\tstatic unsigned char yybm[] = {", o);
 	for(i = 0; b; i += n){
 	    unsigned char m;
 	    unsigned int j;
+	    memset(bm, 0, n);
 	    for(m = 0x80; b && m; b = b->next, m >>= 1){
 		b->i = i; b->m = m;
 		doGen(b->go, b->on, bm-lb, m);
@@ -161,6 +216,7 @@ void BitMap_gen(FILE *o, unsigned int lb, unsigned int ub){
 	    }
 	}
 	fputs("\n\t};\n", o); oline+=2;
+        free(bm);
     }
 }
 
@@ -843,14 +899,17 @@ void DFA_emit(DFA *d, FILE *o){
     nOrgOline = oline;
     maxFillIndexes = vFillIndexes;
     orgVFillIndexes = vFillIndexes;
-    tmpo = fopen("re2c.tmp", "wt");
+#ifdef _WIN32
+    tmpo = win32_tmpfile();
+#else
+    tmpo = tmpfile();
+#endif
     for(s = d->head; s; s = s->next){
 	int readCh = 0;
 	State_emit(s, tmpo, &readCh);
 	Go_genGoto(&s->go, tmpo, s, s->next, &readCh);
     }
     fclose(tmpo);
-    remove("re2c.tmp");
     maxFillIndexes = vFillIndexes;
     vFillIndexes = orgVFillIndexes;
     oline = nOrgOline;
